

set -ex



pip check
export OPENBLAS_NUM_THREADS=1
export OMP_NUM_THREADS=1
export MPLBACKEND=Agg
SKIMAGE_TEST_STRICT_WARNINGS=0 pytest --verbose --pyargs skimage -k "not (_not_a_real_test or test_end_points or test_RGB or test_periodic_reference or test_hdx_rgb_roundtrip or test_hed_rgb_roundtrip or test_apply_parallel_rgb)"
exit 0
